#include "cnn.h"
#include <stdio.h>
#include <pthread.h>

static long total = 0;
pthread_mutex_t mutex;

typedef struct {
  int low;
  int high;
} pthread_arg;

__attribute__((noinline)) void *convolution(void *argument) {

  int om; // output map index
  int im; // input map index
  int row, col; // row, col indices in output map
  int i,j; // filter indices
  
  volatile int NUM_INPUT_MAPS = INPUT_MAPS/2; // prevent LLVM from inlining
  volatile int FILTER_D = FILTER_DIM; // prevent LLVM from inlining
  
  pthread_arg targ = *((pthread_arg *)argument);
  
  for (om = targ.low; om < targ.high; om++) {
    for (row = 0; row < OUTPUT_DIM; row++) {
      for (col = 0; col < OUTPUT_DIM; col++) {
	short output_fm_value = 0;
	for (i = 0; i < FILTER_D; i++) {
	  for (j = 0; j < FILTER_D; j++) {
	    for (im = 0; im < NUM_INPUT_MAPS; im+=2) {
	      // four MACs per loop iteration
	      output_fm_value += weights0[om][i][j][im] * input_fms0[row+i][col+j][im]; 
	      output_fm_value += weights1[om][i][j][im] * input_fms1[row+i][col+j][im];
	      output_fm_value += weights0[om][i][j][im+1] * input_fms0[row+i][col+j][im+1];
              output_fm_value += weights1[om][i][j][im+1] * input_fms1[row+i][col+j][im+1];  
	    }
	  }
	}
	output_fms[row][col][om] = output_fm_value;
	pthread_mutex_lock(&mutex);
	total += output_fm_value; // for correctness checking
	pthread_mutex_unlock(&mutex);
      }
    }
  }
  pthread_exit(NULL);
}

int main(void) {

  pthread_arg arg1, arg2;
  arg1.low = 0;
  arg1.high = 4;
  arg2.low = 4;
  arg2.high = 8;
  
  pthread_t t1, t2;
  pthread_create(&t1, NULL, convolution, &arg1);
  pthread_create(&t2, NULL, convolution, &arg2);

  pthread_join(t1, NULL);
  pthread_join(t2, NULL);
  
  printf("TOTAL: %ld\n", total);
  if (total == -18870)
    printf("PASS\n");
  else
    printf("FAIL\n");
  
  return 0;
}
